/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Analyzer.java 2398 2005-12-02 10:32:29 +0100 (Fri, 02 Dec 2005) michelou $
 */

package zweic;

import zweic.Symbol.*;
import zweic.Tree.*;
import zweic.Type.*;


/**
 * This class implements the semantic analyzer of zweic.
 */
class Analyzer {
    private final ProgramVisitor programVisitor;
    private final ClassVisitor   classVisitor;
    private final TypeVisitor    typeVisitor;
    private final MemberVisitor  memberVisitor;
    private final StatVisitor    statVisitor;
    private final ExprVisitor    exprVisitor;

    // Global scope of classes
    private final ClassScope classScope = new ClassScope();

    public Analyzer() {
        programVisitor = new ProgramVisitor();
        classVisitor   = new ClassVisitor();
        typeVisitor    = new TypeVisitor();
        memberVisitor  = new MemberVisitor();
        statVisitor    = new StatVisitor();
        exprVisitor    = new ExprVisitor();
    }

    /**
     * Analyze the tree
     */
    public Type analyze(Tree tree) {
        programVisitor.analyze(tree);
        return tree.type;
    }

    /**
     * Analyze a program
     */
    private class ProgramVisitor extends DefaultVisitor {

        public void analyze(Tree tree) {
            tree.apply(this);
        }

        public void caseProgram(Program tree) {
            classVisitor.analyzeEach(tree.decls);
            exprVisitor.analyze(new VarScope(), tree.main);
        }

        public void caseDefault(Tree tree) {
            throw new Error("Visitor error in programVisitor");
        }
    }

    /**
     * Analyze a class
     */
    private class ClassVisitor extends DefaultVisitor {

        public void analyze(ClassDef clazz) {
            clazz.apply(this);
        }

        public void analyzeEach(ClassDef[] clazzes) {
            for (int i = 0; i < clazzes.length; i++)
                analyze(clazzes[i]);
        }

        public void caseClassDef(ClassDef tree) {
            // ...  complter ...
        }

        public void caseDefault(Tree tree) {
            throw new Error("Visitor error in classVisitor");
        }
    }

    /**
     * Analyze a type
     */
    private class TypeVisitor extends DefaultVisitor {

        public Type analyze(TypeTree tree) {
            tree.apply(this);
            return tree.type;
        }

        public Type[] analyzeEach(TypeTree[] trees) {
            Type[] types = new Type[trees.length];
            for (int i = 0; i < trees.length; i++)
                types[i] = analyze(trees[i]);
            return types;
        }

        public void caseIntType(IntType tree) {
            tree.type = Type.IIntType;
        }

        public void caseNullType(NullType tree) {
            tree.type = Type.INullType;
        }

        public void caseClassType(ClassType tree) {
            // ...  complter ...
        }

        public void caseDefault(Tree tree) {
            throw new Error("Visitor error in typeVisitor");
        }
    }

    /**
     * Analyze a class member
     */
    private class MemberVisitor extends DefaultVisitor {
        private ClassSymbol ownerClass;

        public void analyze(ClassSymbol ownerClass, Tree member) {
            ClassSymbol oldOwner = this.ownerClass;
            this.ownerClass = ownerClass;
            member.apply(this);
            this.ownerClass = oldOwner;
        }

        public void analyzeEach(ClassSymbol ownerClass, Tree[] members) {
            for (int i = 0; i < members.length; i++)
                analyze(ownerClass, members[i]);
        }

        public void caseFieldDecl(FieldDecl tree) {
            // ...  complter ...
        }

        public void caseMethodDef(MethodDef tree) {
            // ...  complter ...
        }

        public void caseDefault(Tree tree) {
            throw new Error("Visitor error in MemberVisitor");
        }
    }

    /**
     * Analyze a statement
     */
    private class StatVisitor extends DefaultVisitor {
        private VarScope varScope;

        public VarScope analyze(VarScope scope, Stat stat) {
            VarScope oldScope = varScope;
            varScope = new VarScope(scope);
            stat.apply(this);
            scope = varScope;
            varScope = oldScope;
            return scope;
        }

        public VarScope analyzeEach(VarScope scope, Stat[] stats) {
            VarScope currentScope = scope;
            for (int i = 0; i < stats.length; i++)
                currentScope = analyze(currentScope, stats[i]);
            return currentScope;
        }

        public void caseWhile(While tree) {
            checkIntType(tree.cond.pos,
                         exprVisitor.analyze(varScope, tree.cond));
            analyzeEach(varScope, tree.stats);
        }
        // ...  complter ...
        public void caseDefault(Tree tree) {
            throw new Error("Visitor error in StatVisitor");
        }
    }

    /**
     * Analyze an expression
     */
    private class ExprVisitor extends DefaultVisitor {
        private VarScope varScope;

        public Type analyze(VarScope scope, Tree expr) {
            VarScope oldScope = varScope;
            varScope = new VarScope(scope);
            expr.apply(this);
            varScope = oldScope;
            return expr.type;
        }

        public Type[] analyzeEach(VarScope scope, Tree[] exprs) {
            Type[] types = new Type[exprs.length];
            for (int i = 0; i < exprs.length; i++)
                types[i] = analyze(scope, exprs[i]);
            return types;
        }

        public void caseBad(Bad tree) {
            tree.type = Type.IBadType;
        }
        // ...  complter ...
        public void caseDefault(Tree tree) {
            throw new Error("Visitor error in ExprVisitor");
        }
    }

    /**
     * Generates an error if the found type is not equal
     * to the expected type.
     */
    private static void checkSameType(int pos, Type found, Type expected) {
        if (! found.isSameType(expected))
            error(pos, "Type mismatch", found, expected);
    }

    /**
     * Generates an error if the found type is not equal to IIntType.
     */
    private static void checkIntType(int pos, Type found) {
        checkSameType(pos, found, Type.IIntType);
    }

    /**
     * Generates an error if the found type is not a subtype
     * of the expected type.
     */
    private static void checkSubtype(int pos, String header, Type found, Type expected) {
        if (! found.isSubtype(expected))
            error(pos, header, found, expected);
    }

    private static void error(int pos, String header, Type found, Type expected) {
        Report.error(pos, header + "\nexpected type: " + expected +
                          "\nactual type  : " + found);
    }

    /**
     * For the global scope of classes
     */
    private static class ClassScope {
        private final java.util.Map/*<String,ClassSymbol>*/ map;
        ClassScope() { map = new java.util.LinkedHashMap(); }
        void enter(String name, ClassSymbol c) { map.put(name, c); }
        ClassSymbol lookup(String name) { return (ClassSymbol) map.get(name); }
    }

    /**
     * For local scopes of variables
     */
    private static class VarScope {
        private final java.util.Map/*<String,VarSymbol>*/ map;
        VarScope() { map = new java.util.LinkedHashMap(); }
        VarScope(VarScope scope) { map = new java.util.LinkedHashMap(scope.map); }
        void enter(String name, VarSymbol v) { map.put(name, v); }
        VarSymbol lookup(String name) { return (VarSymbol) map.get(name); }
    }

}
