/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Code.java 2698 2006-02-17 18:31:10 +0100 (Fri, 17 Feb 2006) michelou $
 */

package zweic;

import java.io.PrintWriter;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;

public class Code implements RISC {
    //########################################################################
    // Constants

    /** Index of first general-use register. */
    private static final int RC_MIN =  1;
    /** Index of last general-use register. */
    private static final int RC_MAX = 29;

    //########################################################################
    // Instance variables

    /** Free registers */
    private boolean[] freeRegisters;

    /** Already emitted instructions */
    private final List code;

    /** Stack frame size */
    private int frameSize;

    //########################################################################
    // Constructor

    public Code() {
        code = new ArrayList();
        freeRegisters = new boolean[32];
        for (int r = RC_MIN; r <= RC_MAX; ++r)
            freeRegisters[r] = true;
        frameSize = 0;
    }

    //########################################################################
    // Methods - register allocation

    /** Allocate a register. */
    public int getRegister() {
        for (int r = RC_MIN; r <= RC_MAX; ++r)
            if (freeRegisters[r])
                return getRegister(r);
        throw new Error("no more free registers");
    }

    /**
     * Allocate the given register, which has to be free (otherwise an
     * exception is thrown).
     */
    public int getRegister(int r) {
        if (!freeRegisters[r])
            throw new Error("attempt to allocate non-free register R" + r);
        freeRegisters[r] = false;
        return r;
    }

    /** Free a register. */
    public void freeRegister(int r) {
        if (freeRegisters[r])
            throw new Error("attempt to free already-free register R" + r);
        if (r < RC_MIN || r > RC_MAX)
            throw new Error("attempt to free special register R" + r);
        freeRegisters[r] = true;
    }

    /**
     * Return the set of used registers as an array. If element at
     * index I is true, this means that register R[i] is in use.
     */
    public boolean[] usedRegisters() {
        boolean[] used = new boolean[freeRegisters.length];
        for (int i = 0; i < used.length; ++i)
            used[i] = (i >= RC_MIN) && (i <= RC_MAX) && !freeRegisters[i];
        return used;
    }

    //########################################################################
    // Methods - address handling

    /** Return address of next instruction. */
    public int pc() {
        return WORD_SIZE * code.size();
    }

    //########################################################################
    // Methods - frame handling

    /** Increment frame size by given value. */
    public void incFrameSize(int bytes) {
        frameSize += bytes;
    }

    /** Decrement frame size by given value. */
    public void decFrameSize(int bytes) {
        frameSize -= bytes;
        assert frameSize >= 0;
    }

    /** Return current frame size. */
    public int getFrameSize() {
        return frameSize;
    }

    //########################################################################
    // Methods - code emitting

    public void emit(int opcode) {
        emit(opcode, Integer.MIN_VALUE);
    }

    public void emit(int opcode, int c) {
        emit(opcode, Integer.MIN_VALUE, c);
    }

    public void emit(int opcode, int c, String s) {
        emit(opcode, Integer.MIN_VALUE, Integer.MIN_VALUE, c, s);
    }

    public void emit(int opcode, Label l) {
        emit(opcode, Integer.MIN_VALUE, l);
    }

    public void emit(int opcode, Label l, String s) {
        emit(opcode, Integer.MIN_VALUE, l, s);
    }

    public void emit(int opcode, int a, int c) {
        emit(opcode, a, Integer.MIN_VALUE, c);
    }

    public void emit(int opcode, int a, int c, String s) {
        emit(opcode, a, Integer.MIN_VALUE, c, s);
    }

    public void emit(int opcode, int a, Label l) {
        emit(opcode, a, l, null);
    }

    public void emit(int opcode, int a, Label l, String s) {
        if (l.isAnchored())
            emit(opcode, a, (l.getAnchor() - pc()) / WORD_SIZE, s);
        else {
            l.recordInstructionToPatch(pc());
            emit(opcode, a, Integer.MIN_VALUE, Integer.MIN_VALUE, s);
        }
    }

    private static String syscall2string(int c) {
        switch (c) {
        case SYS_IO_FLUSH : return "SYS_IO_FLUSH";
        case SYS_IO_RD_CHR: return "SYS_IO_RD_CHR";
        case SYS_IO_RD_INT: return "SYS_IO_RD_INT";
        case SYS_IO_WR_CHR: return "SYS_IO_WR_CHR";
        case SYS_IO_WR_INT: return "SYS_IO_WR_INT";
        case SYS_GC_INIT  : return "SYS_GC_INIT";
        case SYS_GC_ALLOC : return "SYS_GC_ALLOC";
        case SYS_GET_TOTAL_MEM_SIZE: return "SYS_GET_TOTAL_MEM_SIZE";
        case SYS_EXIT     : return "SYS_EXIT";
        default           : return Integer.toString(c);
        }
    }

    public void emit(int opcode, int a, int b, int c) {
        String s = (opcode == SYSCALL) ? syscall2string(c) : null;
        emit(opcode, a, b, c, s);
    }

    public void emit(int opcode, int a, int b, int c, String s) {
        code.add(new Instruction(opcode, a, b, c, s));
    }

    //########################################################################
    // Methods - program printing.

    /** Print program. */
    public void write(PrintWriter out) {
        for (int i = 0; i < code.size(); i++) {
            String label = Integer.toString(WORD_SIZE * i);
            while (label.length() < 4) label = '0' + label;
            out.print("/* " + label + " */ ");
            out.print(code.get(i));
            out.println();
        }
        out.flush();
    }

    //########################################################################
    // Labels

    public Label getLabel() {
        return new Label();
    }

    public void anchorLabel(Label l) {
        l.setAnchor(pc());
    }

    public class Label {
        private ArrayList/*<Integer>*/ toPatch = new ArrayList();
        private int pc = -1;

        private Label() {}
        private void setAnchor(int pc) {
            this.pc = pc;
            Iterator toPatchIt = toPatch.iterator();
            while (toPatchIt.hasNext()) {
                int instrPC = ((Integer)toPatchIt.next()).intValue();
                ((Instruction)code.get(instrPC / WORD_SIZE)).c =
                    (pc - instrPC) / WORD_SIZE;
            }
        }
        public int getAnchor() { return pc; }
        private boolean isAnchored() { return pc >= 0; }
        private void recordInstructionToPatch(int pc) {
            toPatch.add(new Integer(pc));
        }
    }

    //########################################################################
    // Instructions

    private static class Instruction {

        public final int opcode;
        public final int a;
        public final int b;
        public int c;
        public final String s;

        public Instruction(int opcode, int a, int b, int c, String s) {
            this.opcode = opcode;
            this.a = a;
            this.b = b;
            this.c = c;
            this.s = s;
        }

        public String toString() {
            StringBuffer buffer = new StringBuffer();
            buffer.append(mnemonics[opcode]);
            if (a != Integer.MIN_VALUE) buffer.append(' ').append(a);
            if (b != Integer.MIN_VALUE) buffer.append(' ').append(b);
            if (c != Integer.MIN_VALUE) buffer.append(' ').append(c);
            if (s != null) {
                int i = buffer.length();
                while (i < 17) { buffer.append(' '); i++; }
                buffer.append("/* ").append(s).append(" */");
            }
            return buffer.toString();
        }
    }

    //########################################################################
}
