/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Generator.java 2698 2006-02-17 18:31:10 +0100 (Fri, 17 Feb 2006) michelou $
 */

package zweic;

import zweic.Code.*;
import zweic.Symbol.*;
import zweic.Tree.*;


public class Generator implements RISC {
    private final Gen genVisitor;
    private final GenLoad genLoadVisitor;
    private final GenCond genCondVisitor;
    private final Code code;

    public Generator() {
        genVisitor = new Gen();
        genLoadVisitor = new GenLoad();
        genCondVisitor = new GenCond();
        code = new Code();
    }

    public Code gen(Tree tree) {
        genVisitor.generate(tree);
        return code;
    }

    private void gen(Tree[] trees) {
        for (int i = 0; i < trees.length; ++i)
            genVisitor.generate(trees[i]);
    }

    private void genLoad(Tree tree, int targetReg) {
        assert code.usedRegisters()[targetReg]
            : "invariant violated: target register R" + targetReg
            + " not allocated";
        genLoadVisitor.generate(tree, targetReg);
    }

    private void genCond(Tree tree, Code.Label targetLabel, boolean when) {
        genCondVisitor.generate(tree, targetLabel, when);
    }

    /**
     * Visitor generating code for statements, in a general sense,
     * that is all constructs which do not have a value.
     */
    private class Gen extends DefaultVisitor {

        public void generate(Tree tree) {
            tree.apply(this);
        }

        public void caseProgram(Program tree) {
            Label init = code.getLabel();
            code.emit(BEQ, ZERO, init);

            // ...  complter ...

            // generate class and function definitions
            // ...  complter ...

            // generate main expression
            Label start = code.getLabel();
            code.anchorLabel(start);
            int tmpReg = code.getRegister();
            genLoad(tree.main, tmpReg);
            code.freeRegister(tmpReg);

            // exit
            code.emit(RET, ZERO, "exit program");

            // the following code is placed here because the space that
            // it occupies can be used after initialization.

            code.anchorLabel(init);
            // initialise stack pointer
            code.emit(SYSCALL, SP, 0, SYS_GET_TOTAL_MEM_SIZE); 
            // initialise garbage collector:
            // the heap starts at init and 
            // its size is 2/3 * (total_mem_size - init) words
            int r1 = code.getRegister();
            int r2 = code.getRegister();
            int r3 = code.getRegister();
            emitLoadConstant(r1, init.getAnchor());
            code.emit(SUB, r2, SP, r1); 
            code.emit(DIVIU, r2, r2, 3 * WORD_SIZE);
            code.emit(LSHI, r2, r2, 1); // r2 now contains the size of the heap
            emitLoadConstant(r3, SP << 27);
            code.emit(ADD, r2, r2, r3);
            code.emit(SYSCALL, r1, r2, SYS_GC_INIT);
            code.freeRegister(r3);
            code.freeRegister(r2);
            code.freeRegister(r1);

            // populate VMTs
            // ...  complter ...

            // jump to main expression
            code.emit(BEQ, ZERO, start, "jump to main expression");
        }

        public void caseClassDef(ClassDef tree) {
            // ...  complter ...
        }

        public void caseFieldDecl(FieldDecl tree) {
            // ...  complter ...
        }
        public void caseMethodDef(MethodDef tree) {
            // ...  complter ...
        }

        public void caseWhile(While tree) {
            // ...  complter ...
        }

        public void caseVar(Var tree) {
            // ...  complter ...
        }

        public void caseSet(Set tree) {
            // ...  complter ...
        }

        public void caseDo(Do tree) {
            // ...  complter ...
        }

        public void casePrintInt(PrintInt tree) {
            int tmpReg = code.getRegister();
            genLoad(tree.expr, tmpReg);
            code.emit(SYSCALL, tmpReg, ZERO, SYS_IO_WR_INT, "printInt");
            code.emit(SYSCALL, ZERO, ZERO, SYS_IO_FLUSH, "flush");
            code.freeRegister(tmpReg);
        }

        public void casePrintChar(PrintChar tree) {
            // ...  complter ...
        }

        public void caseDefault(Tree tree) {
            // ...  complter ...
        }
    }

    /**
     * Visitor generating code for expressions, that is constructs
     * which produce a value.
     */
    private class GenLoad extends DefaultVisitor {
        private int targetReg;

        public void generate(Tree tree, int targetReg) {
            int bkpTargetReg = this.targetReg;
            this.targetReg = targetReg;
            tree.apply(this);
            this.targetReg = bkpTargetReg;
        }

        public void caseNullLit(NullLit tree) {
            // ...  complter ...
        }

        public void caseIntLit(IntLit tree) {
            // ...  complter ...
        }

        public void caseReadInt(ReadInt tree) {
            // ...  complter ...
        }

        public void caseReadChar(ReadChar tree) {
            // ...  complter ...
        }

        public void caseUnop(Unop tree) {
            // ...  complter ...
        }

        public void caseBinop(Binop tree) {
            // ...  complter ...
        }

        public void caseBlock(Block tree) {
            // ...  complter ...
        }

        public void caseIf(If tree) {
            // ...  complter ...
        }

        public void caseIdent(Ident tree) {
            // ...  complter ...
        }

        public void caseNew(New tree) {
            // ...  complter ...
        }

        public void caseSelect(Select tree) {
            // ...  complter ...
        }

        public void caseCall(Call tree) {
            // ...  complter ...
        }

        public void caseDefault(Tree tree) {
            // ...  complter ...
        }

    }

    /**
     * Visitor generating code for conditions, that is constructs
     * which jump somewhere on the result of a test.
     */
    private class GenCond extends DefaultVisitor {
        private Code.Label targetLabel;
        private boolean when;

        public void generate(Tree tree, Code.Label targetLabel, boolean when) {
            Code.Label bkpTargetLabel = this.targetLabel;
            this.targetLabel = targetLabel;
            boolean bkpWhen = this.when;
            this.when = when;
            tree.apply(this);
            this.targetLabel = bkpTargetLabel;
            this.when = bkpWhen;
        }

        public void caseBinop(Binop tree) {
            // ...  complter ...
        }

        public void caseUnop(Unop tree) {
            // ...  complter ...
        }

        public void caseDefault(Tree tree) {
            // ...  complter ...
        }

    }

}
