/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Printer.java 2352 2005-11-24 09:53:21 +0100 (Thu, 24 Nov 2005) michelou $
 */

package zweic;

import java.io.PrintWriter;
import zweic.Tree.*;


/**
 * This class implements a pretty printer for abstract syntax
 * trees in Zwei.
 */
public class Printer implements Visitor {

    /**
     * Indentation string.
     */
    private String step = "  ";

    /**
     * indentation level.
     */
    private int level = 0;

    /**
     * printer output.
     */
    private PrintWriter out;

    Printer(PrintWriter out) {
        this.out = out;
    }

    /**
     * Print a tree node.
     */
    public Printer print(Tree tree) {
        tree.apply(this);
        return this;
    }

    /**
     * Print abstract syntax trees separated by commas.
     */
    private Printer print(Tree[] trees) {
        for(int i = 0; i < trees.length; i++) {
            if (i > 0) print(", ");
            print(trees[i]);
        }
        return this;
    }

    /**
     * Print a number.
     */
    private Printer print(int value) {
        return print(String.valueOf(value));
    }

    /**
     * Print a string.
     */
    private Printer print(String value) {
        out.print(value);
        return this;
    }

    /**
     * Print an identifier.
     */
    private Printer print(Name ident) {
        return print(ident.name);
    }

    /**
     * Print a formal parameter.
     */
    private Printer print(Formal param) {
        return print(param.typeName).print(" ").print(param.name);
    }

    /**
     * Print formal parameters separated by commas.
     */
    private Printer print(Formal[] params) {
        for(int i = 0; i < params.length; i++) {
            if (i > 0) print(", ");
            print(params[i]);
        }
        return this;
    }

    /**
     * Print an end-of-line character and indent the following
     * line appropriately.
     */
    private Printer println() {
        out.println();
        for (int i = 0; i < level; i++)
            out.print(step);
        return this;
    }
    
    /**
     * Increment the indentation level.
     */
    private Printer indent() {
        level++;
        return this;
    }
    
    /**
     * Decrement the indentation level.
     */
    private Printer undent() {
        level--;
        return this;
    }

    //// VISITOR IMPLEMENTATION ////

    public void caseBad(Bad tree) {
        print("<bad>");
    }

    public void caseProgram(Program tree) {
        for (int i = 0; i < tree.decls.length; i++)
            print(tree.decls[i]).println();
        print(tree.main).println();
    }
    // ...  complter ...
    public void caseSelect(Select tree) {
        print(tree.prefix).print(".").print(tree.selector);
    }
}
