/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Symbol.java 2241 2005-11-18 16:41:27 +0100 (Fri, 18 Nov 2005) michelou $
 */

package zweic;

import java.util.LinkedHashMap; // preserves insertion order !!!


/**
 * This class provides a representation for symbol table entries.
 */
abstract class Symbol {

    /**
     * The position of the symbol's definition.
     */
    public int pos;

    /**
     * The name of the symbol.
     */
    public String name;
   
    /**
     * Offset of the symbol. For variables, that's the offset
     * relative to the frame pointer; for functions that's the PC of
     * the first instruction.
     */
    public int offset;

    public static class ClassSymbol extends Symbol {
        private final ClassSymbol superclass;
        private final LinkedHashMap/*<String,FieldSymbol>*/ fields;
        private final LinkedHashMap/*<String,MethodSymbol>*/ methods;

        ClassSymbol(int pos, String name, ClassSymbol superclass) {
            this.pos = pos;
            this.name = name;
            this.superclass = superclass;
            this.fields = (superclass != null) ? new LinkedHashMap(superclass.fields) : new LinkedHashMap();
            this.methods = (superclass != null) ? new LinkedHashMap(superclass.methods) : new LinkedHashMap();
        }

        public ClassSymbol[] getSuperClasses() {
           if (superclass != null) {
               ClassSymbol[] s = superclass.getSuperClasses();
               ClassSymbol[] s1 = new ClassSymbol[s.length + 1];
               s1[0] = this;
               System.arraycopy(s, 0, s1, 1, s.length);
               return s1;
           }
           else
               return new ClassSymbol[]{ this };
        }
    
        public boolean isSubclassOf(ClassSymbol that) {
            ClassSymbol[] s = getSuperClasses();
            for (int i = 0; i < s.length; i++)
                if (s[i] == that) return true;
            return false;
        }
    
        public void enterField(FieldSymbol f) {
            f.offset = fields.size() * 4;
            fields.put(f.name, f);
        }
    
        public FieldSymbol lookupField(String name) {
            return (FieldSymbol) fields.get(name);
        }
    
        public void enterNewMethod(MethodSymbol m) {
            m.offset = methods.size() * 4;
            methods.put(m.name, m);
        }

        public void enterMethod(MethodSymbol m) {
            MethodSymbol m1 = (MethodSymbol) methods.get(m.name);
            assert m1 != null : "method \"" + m.name + "\" not found";
            m.offset = m1.offset;
            methods.put(m.name, m);
        }

        public MethodSymbol lookupMethod(String name) {
            return (MethodSymbol) methods.get(name);
        }

        public FieldSymbol[] allFields() {
           FieldSymbol[] all = new FieldSymbol[fields.size()];
           return (FieldSymbol[]) fields.values().toArray(all);
        }

        public Type[] allFieldTypes() {
           FieldSymbol[] all = allFields();
           Type[] allTypes = new Type[all.length];
           for (int i = 0; i < all.length; i++)
               allTypes[i] = all[i].fieldtype;
           return allTypes;
        }

        public MethodSymbol[] allMethods() {
           MethodSymbol[] all = new MethodSymbol[methods.size()];
           return (MethodSymbol[]) methods.values().toArray(all);
        }
    
        public String toString() { return "class " + name; }
    }
    
    public static class FieldSymbol extends Symbol {
        Type fieldtype;
    
        FieldSymbol(int pos, String name, Type fieldtype) {
            this.pos = pos;
            this.name = name;
            this.fieldtype = fieldtype;
        }
    
        public String toString() { return "field " + name + ": " + fieldtype; }
    }
    
    public static class MethodSymbol extends Symbol {
        Type[] paramtypes;
        Type restype;
        int address;
    
        MethodSymbol(int pos, String name, Type[] paramtypes, Type restype) {
            this.pos = pos;
            this.name = name;
            this.paramtypes = paramtypes;
            this.restype = restype;
        }
    
        public String toString() { return "method " + name + "(..): " + restype; }
    }
    
    public static class VarSymbol extends Symbol {
        Type vartype;
    
        VarSymbol(int pos, String name, Type vartype) {
            this.pos = pos;
            this.name = name;
            this.vartype = vartype;
        }
    
        public String toString() { return "variable " + name + ": " + vartype; }
    }

}
