/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Tree.java 2325 2005-11-22 11:16:18 +0100 (Tue, 22 Nov 2005) michelou $
 */

package zweic;


/**
 * Class <code>Tree</code> defines the type of all abstract
 * syntax tree nodes.
 */
abstract class Tree {

    /**
     * The source code position of this tree node.
     */
    public int pos = Position.UNDEFINED;

    public Tree(int pos) { this.pos = pos; }


    /**
     * Applies the given visitor object to this tree node.
     */
    public abstract void apply(Visitor v);
    
    /**
     * This node represents an erroneous syntactical construct.
     */
    public static class Bad extends Tree {
        public Bad(int pos) { super(pos); }
        public void apply(Visitor v) { v.caseBad(this); }
    }

    /**
     * P = Program { D } E
     */
    public static class Program extends Tree {
        public ClassDef[] decls;
        public Tree main;
        
        public Program(int pos, ClassDef[] decls, Tree main) {
            super(pos);
            this.decls = decls;
            this.main = main;
        }

        public void apply(Visitor v) { v.caseProgram(this); }
    }
    
    /**
     * D = ClassDef ident { M }
     */
    public static class ClassDef extends Tree {
    	// ...  complter ...
        public void apply(Visitor v) { v.caseClassDef(this); }
    }

    /**
     * M = FieldDecl name T
     */
    public static class FieldDecl extends Tree {
    	// ...  complter ...
        public void apply(Visitor v) { v.caseFieldDecl(this); }
    }

    /**
     * M = MethodDef name { F } T E
     */
    public static class MethodDef extends Tree {
        public final Name self = new Name("this");
    	// ...  complter ...
        public void apply(Visitor v) { v.caseMethodDef(this); }
    }

    //########################################################################
    // Tree nodes for types

    /**
     * A common superclass for tree nodes designating types
     */
    public static abstract class TypeTree extends Tree {
        public TypeTree(int pos) { super(pos); }
    }

    /**
     * T = Int
     */
    public static class IntType extends TypeTree {
        public IntType(int pos) { super(pos); }
        public void apply(Visitor v) { v.caseIntType(this); }
    }

    /**
     * T = Null
     */
    public static class NullType extends TypeTree {
        public NullType(int pos) { super(pos); }
        public void apply(Visitor v) { v.caseNullType(this); }
    }

    /**
     * T = ClassType name
     */
    public static class ClassType extends TypeTree {
        public Name name;

        public ClassType(int pos, Name name) {
            super(pos);
            this.name = name;
        }
        
        public void apply(Visitor v) { v.caseClassType(this); }
    }

    //########################################################################
    // Tree nodes for statements

    /**
     * A common superclass for all statements
     */
    protected static abstract class Stat extends Tree {
        public Stat(int pos) { super(pos); }
    }

    /**
     * S = While E { S }
     */
    public static class While extends Stat {
        public Tree cond;
        public Stat[] stats;

        public While(int pos, Tree cond, Stat[] stats) {
            super(pos);
            this.cond = cond;
            this.stats = stats;
        }

        public void apply(Visitor v) { v.caseWhile(this); }
    }

    /**
     * S = Var name F E
     */
    public static class Var extends Stat {
        public Name name;
        public TypeTree vartype;
        public Tree init;

        public Var(int pos, Name name, TypeTree vartype, Tree init) {
            super(pos);
            this.name = name;
            this.vartype = vartype;
            this.init = init;
        }
        
        public void apply(Visitor v) { v.caseVar(this); }
    }

    /**
     * S = Set name E
     */
    public static class Set extends Stat {
        public Name name;
        public Tree expr;

        public Set(int pos, Name name, Tree expr) {
            super(pos);
            this.name = name;
            this.expr = expr;
        }

        public void apply(Visitor v) { v.caseSet(this); }
    }

    /**
     * S = Do E
     */
    public static class Do extends Stat {
        public Tree expr;

        public Do(int pos, Tree expr) {
            super(pos);
            this.expr = expr;
        }

        public void apply(Visitor v) { v.caseDo(this); }
    }

    /**
     * S = PrintInt E
     */
    public static class PrintInt extends Stat {
        public Tree expr;

        public PrintInt(int pos, Tree expr) {
            super(pos);
            this.expr = expr;
        }

        public void apply(Visitor v) { v.casePrintInt(this); }
    }

    /**
     * S = PrintChar E
     */
    public static class PrintChar extends Stat {
        public Tree expr;

        public PrintChar(int pos, Tree expr) {
            super(pos);
            this.expr = expr;
        }

        public void apply(Visitor v) { v.casePrintChar(this); }
    }

    //########################################################################
    // Tree nodes for expressions

    // ...  complter ...

    /**
     * B = Add | Sub | Mul | Div | Mod
     *   | Eq | Ne | Lt | Le | Gt | Ge | Or
     */
    public static final int ADD =  0;
    public static final int SUB =  1;
    public static final int OR  =  2;
    public static final int MUL =  3;
    public static final int DIV =  4;
    public static final int MOD =  5;
    public static final int AND =  6;
    public static final int EQ  =  7;
    public static final int NE  =  8;
    public static final int LT  =  9;
    public static final int LE  = 10;
    public static final int GT  = 11;
    public static final int GE  = 12;

    /**
     * E = Unop U E
     */
    public static class Unop extends Tree {
        public int op;
        public Tree expr;

        public Unop(int pos, int op, Tree expr) {
            super(pos);
            this.op = op;
            this.expr = expr;
        }
        
        public void apply(Visitor v) { v.caseUnop(this); }
    }

    /**
     * U = Not | Neg
     */
    public static final int NOT =  13;
    public static final int NEG =  14;

    // ...  complter ...
}
